#include "lines_of_action.hpp"
#include <sstream>

using namespace std;

namespace game
{
  static vector<vector<uint64_t>> create_hash_values()
  {
    default_random_engine generator;
    uniform_int_distribution<uint64_t> distribution;
    vector<vector<uint64_t>> res(8, vector<uint64_t>(8, 0));
    for (int i = 0; i < 8; ++i) 
      {
	for (int j = 0; j < 8; ++j)
	  {
	    res[i][j] = distribution(generator);
	  }
      }
    return res;
  }

  vector<vector<uint64_t>> lines_of_action::black_hash_values = create_hash_values();
  vector<vector<uint64_t>> lines_of_action::white_hash_values = create_hash_values();

  shared_ptr<game<lines_of_action_state>> lines_of_action::do_copy() const
  {
    return shared_ptr<lines_of_action>(new lines_of_action(*this));
  }

  lines_of_action_state lines_of_action::get_state()
  {
    return state;
  }
  
  void lines_of_action::set_state(const lines_of_action_state& s)
  {
    state = s; 
  }

  bool lines_of_action::end_of_game() const
  {
    return state.first_player_win || state.second_player_win;
  }

  bool lines_of_action::won(std::uint8_t player) const
  {
    if (player == 0) return state.first_player_win;
    return state.second_player_win;
  }
  
  bool lines_of_action::lost(std::uint8_t player) const
  {
    if (player == 0) return state.second_player_win;
    return state.first_player_win;
  }

  bool lines_of_action::draw(std::uint8_t player) const
  {
    return false;
  }

  uint8_t lines_of_action::current_player() const
  {
    return state.depth & 1 ? WHITE : BLACK;
  }

  int lines_of_action::value(uint8_t player) const
  {
    if (player == BLACK)
      {
	return state.first_player_win ? 1 : (state.second_player_win ? -1 : 0);
      }
    else if (player == WHITE)
      {
	return state.second_player_win ? 1 : (state.first_player_win ? -1 : 0);
      }
    return 0;
  }

  uint16_t lines_of_action::number_of_moves() const
  {
    return state.nb_moves;
  }

#define coord2bit(i, j) (1LL << (((i)<<3) + (j)))

  uint8_t lines_of_action::get(uint8_t i, uint8_t j) const
  {
    if (state.black_bitboard & coord2bit(i, j)) return BLACK;
    if (state.white_bitboard & coord2bit(i, j)) return WHITE;
    return EMPTY;
  }

#define is_set(bitboard, i, j) (bitboard & coord2bit(i, j))

  bool lines_of_action::is_valid_row_right(uint64_t bitboard1, uint64_t bitboard2, uint8_t from_i, uint8_t from_j, uint8_t n) const
  {
    uint8_t dest_j = from_j + n;
    if (dest_j > 7) return false;
    if (is_set(bitboard1, from_i, dest_j)) return false;
    for (int j = from_j + 1; j < dest_j; ++j)
      {
	if (is_set(bitboard2, from_i, j)) return false;
      }
    return true;
  }

  bool lines_of_action::is_valid_row_left(uint64_t bitboard1, uint64_t bitboard2, uint8_t from_i, uint8_t from_j, uint8_t n) const
  {
    if (n > from_j) return false;
    uint8_t dest_j = from_j - n;
    if (is_set(bitboard1, from_i, dest_j)) return false;
    for (int j = from_j - 1; j > dest_j; --j)
      {
	if (is_set(bitboard2, from_i, j)) return false;
      }
    return true;
  }

  bool lines_of_action::is_valid_column_up(uint64_t bitboard1, uint64_t bitboard2, uint8_t from_i, uint8_t from_j, uint8_t n) const
  {
    if (n > from_i) return false;
    uint8_t dest_i = from_i - n;
    if (is_set(bitboard1, dest_i, from_j)) return false;
    for (int i = from_i - 1; i > dest_i; --i)
      {
	if (is_set(bitboard2, i, from_j)) return false;
      }
    return true;
  }

  bool lines_of_action::is_valid_column_down(uint64_t bitboard1, uint64_t bitboard2, uint8_t from_i, uint8_t from_j, uint8_t n) const
  {
    uint8_t dest_i = from_i + n;
    if (dest_i > 7) return false;
    if (is_set(bitboard1, dest_i, from_j)) return false;
    for (int i = from_i + 1; i < dest_i; ++i)
      {
	if (is_set(bitboard2, i, from_j)) return false;
      }
    return true;
  }

  void lines_of_action::update_state(const move& m)
  {

  }

  void lines_of_action::play(uint16_t m)
  {    
    update_state(move_index_to_move(m));
  }

  string lines_of_action::player_to_string(uint8_t player) const
  {
    return player == BLACK ? "@" : (player == WHITE ? "O" : " ");
  }

  string lines_of_action::move_to_string(uint16_t m) const
  {
    return move_index_to_move(m).to_string();
  }

  string lines_of_action::to_string() const
  {
    stringbuf buffer;
    ostream os(&buffer);
    os << "  a b c d e f g h" << endl;
    for (int i = 0; i < 8; ++i) 
      {
	os << " ";
	for (int k = 0; k < 8; ++k) os << "+-";
	os << "+" << endl;
	os << 8 - i << "|";
	for (int j = 0; j < 8; ++j) 
	  {
	    os << player_to_string(get(i, j)) << "|";
	  }
	os << 8 - i << endl;
      }
    os << " ";
    for (int k = 0; k < 8; ++k) os << "+-";
    os << "+" << endl;
    os << "  a b c d e f g h" << endl;
    return buffer.str();
  }

  void lines_of_action::playout(mt19937& engine, int max_depth)
  {
    while (state.depth <= max_depth && !end_of_game())
      {
  	uniform_int_distribution<uint16_t> distribution(0, number_of_moves() - 1);
  	uint16_t move = distribution(engine);
  	play(move);
      }
  }

  std::uint64_t lines_of_action::hash() const
  {
    return state.hash_value;
  }

  std::uint64_t lines_of_action::hash(std::uint16_t m) const
  {
    return 0;
  }
  
  ostream& operator<<(ostream& os, const lines_of_action& loa)
  {
    os << loa.to_string() << endl;
    return os;
  }
}
